use std::{
	array,
	borrow::{Borrow, BorrowMut},
	collections::{HashMap, VecDeque},
	ops::{Deref, DerefMut},
	rc::Rc,
};

use super::parse::ast;

pub fn eval(program: &ast::Program) {
	let mut realm = Realm::default();
	realm.eval_block(&program.body);

	for stat in &program.body.statements {
		realm.eval_statement(stat);
	}
}

#[derive(Debug, Default)]
pub struct Realm {
	scope: Scope<'static>,
}

#[derive(Debug, Default)]
pub struct Scope<'p> {
	parent: Option<Box<&'p mut Scope<'p>>>,
	locals: HashMap<ast::Name, Value>,
}

impl<'p> Scope<'p> {
	fn child<'c: 'p>(&'c mut self) -> Scope<'c> {
		Scope {
			parent: Some(Box::new(self)),
			..Default::default()
		}
	}

	fn find_name(&mut self, name: &ast::Name) -> Option<&mut Value> {
		let own = self.locals.get_mut(name);
		own.or_else(|| self.parent.as_mut().map(|p| p.find_name(name)).flatten())
	}

	fn set(&mut self, name: ast::Name, value: Value) {
		self.locals.insert(name, value);
	}

	fn update(&mut self, name: &ast::Name, mut value: Value) -> bool {
		let mut found = self.locals.get_mut(name);
		if found.is_none() {
			return false;
		}
		found.insert(&mut value);
		true
	}
}

impl Realm {
	fn eval_block(&mut self, block: &ast::Body) -> RealmResult<Option<Value>> {
		for stat in &block.statements {
			self.eval_statement(stat)?;
		}
		let result = block.last_expr.as_ref().map(|e| self.eval_expr(&e.0));
		match result {
			Some(Result::Err(error)) => Err(error),
			Some(Result::Ok(result)) => Ok(Some(result)),
			None => Ok(None),
		}
	}

	fn eval_statement(&mut self, stat: &ast::Statement) -> RealmResult<()> {
		match stat {
			ast::Statement::Definition(def) => self.eval_def(def),
			ast::Statement::Assign(assign) => self.eval_assign(&assign),
			ast::Statement::Expression(expr) => Ok(drop(self.eval_expr(&expr))),
		}
	}

	fn eval_def(&mut self, def: &ast::Definition) -> RealmResult<()> {
		let value = self.eval_expr(&def.value)?;

		fn rec(scope: &mut Scope, pattern: &ast::Pattern, value: &Value) -> RealmResult<()> {
			match pattern {
				ast::Pattern::Name(name) => scope.set(name.clone(), value.clone()),
				ast::Pattern::TuplDestr(tup) => {
					let Value::Array(arr) = value else {
						Err(RealmError::IllegalAccess)?
					};
					if arr.len() != tup.fields.len() {
						Err(RealmError::IllegalAccess)?
					}
					for (pattern, value) in tup.fields.iter().zip(arr.iter()) {
						rec(scope, pattern, value)?
					}
				}
				ast::Pattern::ArrDestr(arr_pat) => {
					let Value::Array(arr) = value else {
						Err(RealmError::IllegalAccess)?
					};
					let expected_len = arr_pat.item_bef.len() + arr_pat.item_aft.len();
					let fold_len = arr.len() - expected_len;
					if arr.len() < expected_len {
						Err(RealmError::IllegalAccess)?
					}
					let mut values = arr.iter();
					for (pattern, value) in arr_pat.item_bef.iter().zip(&mut values) {
						rec(scope, pattern, value)?
					}
					if let Some(fold) = &arr_pat.item_fold {
						let value = (&mut values).take(fold_len).cloned().collect();
						let value = Value::new_array(value);
						rec(scope, fold, &value)?
					}
					for (pattern, value) in arr_pat.item_aft.iter().zip(values) {
						rec(scope, pattern, value)?
					}
				}
				ast::Pattern::ObjDestr(obj_pat) => {
					let Value::Obj(obj) = value else {
						Err(RealmError::IllegalAccess)?
					};
					for (name, pattern) in &obj_pat.fields {
						let Some(value) = obj.get(&name.0) else {
							Err(RealmError::IllegalAccess)?
						};
						rec(scope, pattern, value)?
					}
				}
			};

			Ok(())
		}

		rec(&mut self.scope, &def.defined, &value);
		Ok(())
	}

	fn eval_assign(&mut self, assign: &ast::Assign) -> RealmResult<()> {
		let value = self.eval_expr(&assign.value);
		todo!()
	}

	fn eval_expr(&mut self, expr: &ast::Expression) -> RealmResult<Value> {
		todo!()
	}
}

#[derive(Debug, Clone, Copy)]
pub enum RealmError {
	IllegalAccess,
}
#[must_use]
pub type RealmResult<T> = Result<T, RealmError>;

pub type BoxVal = Box<Value>;
pub type Object = Rc<HashMap<String, Value>>;
pub type Array = Rc<Vec<Value>>;

#[derive(Debug, Clone)]
pub enum Value {
	Bool(bool),
	Int(i32),
	Float(f32),
	String(String),
	Obj(Object),
	Array(Array),
}

pub fn error(message: impl Into<String>) -> Value {
	Value::Obj(Rc::new(HashMap::from([(
		"message".into(),
		Value::String(message.into()),
	)])))
}

impl Value {
	fn new_array(values: Vec<Value>) -> Self {
		Self::Array(Rc::new(values))
	}

	fn index(&self, key: &Self) -> Result<Self, Self> {
		let result = match (self, key) {
			(Self::Array(arr), Self::Int(num)) => (*num >= 0).then(|| arr.get(*num as usize)).flatten(),
			(Self::Obj(obj), Self::String(str)) => obj.get(str),
			_ => None,
		};

		let Some(result) = result else {
			Err(error("Illegal access"))?
		};

		Ok(result.clone())
	}
}
