use crate::lib::{
	parse::{token::Tok, AstParser},
	utils::Boxable,
};

use std::{
	any::{type_name, type_name_of_val},
	collections::HashMap,
	fmt::Display,
};

use chumsky::{prelude::*, Parser as _};
use text::{ident, int, whitespace};

use super::token::Key;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Program {
	pub body: Body,
}

impl Program {
	pub fn parser() -> impl AstParser<Self> {
		let expr = Expression::parser();
		Body::parser(expr).map(|body| Self { body }).then_ignore(end())
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Statement {
	Definition(Definition),
	Assign(Assign),
	Expression(Expression),
}

impl Statement {
	fn parser(expr: impl AstParser<Expression> + 'static) -> impl AstParser<Self> {
		let def = Definition::parser(expr.clone()).map(Self::Definition);
		let assign = Assign::parser(expr).map(Self::Assign);
		let expr = Expression::parser().map(Self::Expression);
		def.or(assign).or(expr)
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Definition {
	pub defined: Pattern,
	pub value: Expression,
}

impl Definition {
	fn parser(expr: impl AstParser<Expression> + 'static) -> impl AstParser<Self> {
		just(Tok::Let)
			.ignore_then(Pattern::parser())
			.then_ignore(just(Tok::Key(Key::Assign)))
			.then(expr)
			.map(|(defined, value)| Self { defined, value })
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Assign {
	pub assigned: Pattern,
	pub value: Expression,
}

impl Assign {
	fn parser(expr: impl AstParser<Expression> + 'static) -> impl AstParser<Self> {
		Pattern::parser()
			.then_ignore(just(Tok::Key(Key::Assign)))
			.then(expr)
			.map(|(assigned, value)| Self { assigned, value })
			.boxed()
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Name(pub String);

impl Name {
	fn parser() -> impl AstParser<Self> {
		select! { Tok::Id(name) => name.clone() }.map(Self)
	}
}

pub fn name(name: impl ToString) -> Name {
	Name(name.to_string())
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Pattern {
	Name(Name),
	TuplDestr(TuplDestr),
	ArrDestr(ArrDestr),
	ObjDestr(ObjDestr),
	// todo : decide if Index(Expression),
}

impl Pattern {
	fn parser() -> impl AstParser<Self> {
		Name::parser().map(Self::Name)
	}

	pub fn as_accessors(&self) -> Vec<(Name, Vec<Access>)> {
		todo!()
	}
}

#[test]
fn test_property_chain() {
	let pattern = Pattern::ObjDestr(ObjDestr {
		fields: vec![
			(name("a"), Pattern::Name(name("a"))),
			(name("b"), Pattern::Name(name("c"))),
			(
				name("d"),
				Pattern::TuplDestr(TuplDestr {
					fields: vec![Pattern::Name(name("e")), Pattern::Name(name("f"))],
				}),
			),
			(name("g"), Pattern::TuplDestr(TuplDestr { fields: vec![] })),
			(
				name("h"),
				Pattern::ArrDestr(ArrDestr {
					item_bef: vec![Pattern::Name(name("i")), Pattern::Name(name("j"))],
					item_fold: None,
					item_aft: vec![],
				}),
			),
			(
				name("k"),
				Pattern::ArrDestr(ArrDestr {
					item_bef: vec![Pattern::Name(name("l"))],
					item_fold: Some(Pattern::Name(name("m")).into_box()),
					item_aft: vec![Pattern::Name(name("n")), Pattern::Name(name("o"))],
				}),
			),
			(name("p"), Pattern::ObjDestr(ObjDestr { fields: vec![] })),
		],
	});
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct TuplDestr {
	pub fields: Vec<Pattern>,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct ArrDestr {
	pub item_bef: Vec<Pattern>,
	pub item_fold: Option<Box<Pattern>>,
	pub item_aft: Vec<Pattern>,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct ObjDestr {
	pub fields: Vec<(Name, Pattern)>,
}

impl ObjDestr {
	fn parser() -> impl AstParser<Self> {
		todo()
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Expression {
	Literal(Literal),
	Variable(Name),
	Block(Body),
	Op(BoxEx, Op, BoxEx),
	Access(BoxEx, Access),
}

impl Expression {
	#[allow(clippy::let_and_return)]
	fn parser() -> impl AstParser<Self> {
		recursive(|rec| {
			let literal = Literal::parser().map(Self::Literal);
			let variable = Name::parser().map(Self::Variable);
			let value = literal.or(variable);

			let closed = value.or(rec
				.clone()
				.boxed()
				.delimited_by(just(Tok::Key(Key::BracketsL)), just(Tok::Key(Key::BracketsR))));

			let access = closed
				.then(Access::parser(rec.clone()).repeated())
				.foldl(|e, a| Self::Access(BoxEx::new(e), a));

			let product = access
				.clone()
				.then(Op::parser_mul().then(access).repeated())
				.foldl(|a, (op, b)| Self::Op(BoxEx::new(a), op, BoxEx::new(b)));

			let addition = product
				.clone()
				.then(Op::parser_add().then(product).repeated())
				.foldl(|a, (op, b)| Self::Op(BoxEx::new(a), op, BoxEx::new(b)));

			let comparisons = addition
				.clone()
				.then(Op::parser_cmp().then(addition).repeated())
				.foldl(|a, (op, b)| Self::Op(BoxEx::new(a), op, BoxEx::new(b)));

			comparisons
		})
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct BoxEx(pub Box<Expression>);

impl BoxEx {
	fn new(expr: Expression) -> Self {
		Self(Box::new(expr))
	}
}

trait ExprParser<I: Clone>: Parser<I, Expression> {
	fn boxify(self) -> chumsky::combinator::Map<Self, fn(Expression) -> BoxEx, Expression>
	where
		Self: Sized,
	{
		self.map(|e| BoxEx(Box::new(e)))
	}
}
impl<T, I> ExprParser<I> for T
where
	T: Parser<I, Expression>,
	I: Clone,
{
}

#[derive(Debug, Clone, PartialEq)]
pub enum Literal {
	Number(f32),
	String(String),
	Bool(bool),
}

impl std::hash::Hash for Literal {
	fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
		core::mem::discriminant(self).hash(state);
	}
}

impl Eq for Literal {}

fn parse_num(num: &str) -> Option<f32> {
	if num.starts_with('.') {
		format!("0{num}").parse().ok()
	} else {
		num.parse().ok()
	}
}

impl Literal {
	fn parser() -> impl AstParser<Self> {
		let num = select! { Tok::Num(num) => num }
			.map(|s| parse_num(&s).unwrap())
			.map(Self::Number);
		let str = select! {Tok::Str(str) => str}.map(Self::String);
		let bool = select! { Tok::Bool(bool) => bool }.map(Self::Bool);
		num.or(str).or(bool)
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Access {
	Invoke(Invoke),
	Index(Index),
	Property(Property),
}

impl Access {
	fn parser(expr: impl AstParser<Expression>) -> impl AstParser<Self> {
		Invoke::parser(expr.clone())
			.map(Self::Invoke)
			.or(Index::parser(expr).map(Self::Index))
			.or(Property::parser().map(Self::Property))
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Invoke {
	pub arguments: Vec<Expression>,
}

impl Invoke {
	fn parser(expr: impl AstParser<Expression>) -> impl AstParser<Self> {
		expr.separated_by(just(Tok::Key(Key::Comma)))
			.delimited_by(just(Tok::Key(Key::BracketsL)), just(Tok::Key(Key::BracketsR)))
			.map(|arguments| Self { arguments })
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Index {
	pub index: BoxEx,
}

impl Index {
	fn parser(expr: impl AstParser<Expression>) -> impl AstParser<Self> {
		expr.delimited_by(just(Tok::Key(Key::SquBrackL)), just(Tok::Key(Key::SquBrackR)))
			.boxify()
			.map(|index| Self { index })
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Property {
	pub name: Name,
}

impl Property {
	fn parser() -> impl AstParser<Self> {
		Name::parser().map(|index| Self { name: index })
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Op {
	Add,
	Sub,
	Div,
	Mul,
	Eq,
	Neq,
	Lt,
	Gt,
	Leq,
	Geq,
	And,
	Or,
}

impl Op {
	fn parser_add() -> impl AstParser<Self> {
		let add = just(Tok::Key(Key::Add)).to(Self::Add);
		let sub = just(Tok::Key(Key::Sub)).to(Self::Sub);
		add.or(sub)
	}
	fn parser_mul() -> impl AstParser<Self> {
		let div = just(Tok::Key(Key::Div)).to(Self::Div);
		let mul = just(Tok::Key(Key::Mul)).to(Self::Mul);
		div.or(mul)
	}

	fn parser_cmp() -> impl AstParser<Self> {
		let eq = just(Tok::Key(Key::Eq)).to(Self::Eq);
		let neq = just(Tok::Key(Key::Neq)).to(Self::Neq);
		let lt = just(Tok::Key(Key::AngBrackL)).to(Self::Lt);
		let gt = just(Tok::Key(Key::AngBrackR)).to(Self::Gt);
		let leq = just(Tok::Key(Key::Leq)).to(Self::Leq);
		let geq = just(Tok::Key(Key::Geq)).to(Self::Geq);
		eq.or(neq).or(lt).or(gt).or(leq).or(geq)
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Body {
	pub statements: Vec<Statement>,
	pub last_expr: Option<BoxEx>,
}

impl Body {
	fn parser(expr: impl AstParser<Expression> + 'static) -> impl AstParser<Self> {
		let statements = Statement::parser(expr.clone())
			.then_ignore(just(Key::Semi.tok()))
			.repeated();
		statements
			.then(expr.boxify().or_not())
			.map(|(statements, last_expr)| Self { statements, last_expr })
	}
}
