use std::fmt::Display;

use chumsky::{prelude::*, Parser as _};
use text::{ident, int, whitespace};

use crate::lib::Span;

use super::AnyParser;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Key {
	/// `!`
	Not,
	/// `-`
	Sub,
	/// `+`
	Add,
	/// `*`
	Mul,
	/// `/`
	Div,
	/// `-=`
	SubEq,
	/// `+=`
	AddEq,
	/// `*=`
	MulEq,
	/// `/=`
	DivEq,
	/// `==`
	Eq,
	/// `!=`
	Neq,
	/// `<=`
	Leq,
	/// `>=`
	Geq,
	/// `(`
	BracketsL,
	/// `)`
	BracketsR,
	/// `{`
	CurBrackL,
	/// `}`
	CurBrackR,
	/// `[`
	SquBrackL,
	/// `]`
	SquBrackR,
	/// `<`
	AngBrackL,
	/// `>`
	AngBrackR,
	/// `:`
	Colon,
	/// `;`
	Semi,
	/// `,`
	Comma,
	/// `.`
	Dot,
	/// `=`
	Assign,
	/// `...`
	Unfold,
}

impl Key {
	pub fn parser() -> impl AnyParser<char, Self> {
		one_of("")
			.to(Key::Add)
			.or(just('!').to(Key::Not))
			.or(just("+=").to(Key::AddEq))
			.or(just("-=").to(Key::SubEq))
			.or(just("*=").to(Key::MulEq))
			.or(just("/=").to(Key::DivEq))
			.or(just('-').to(Key::Add))
			.or(just('+').to(Key::Sub))
			.or(just('*').to(Key::Mul))
			.or(just('/').to(Key::Div))
			.or(just("<=").to(Key::Leq))
			.or(just(">=").to(Key::Geq))
			.or(just('(').to(Key::BracketsL))
			.or(just(')').to(Key::BracketsR))
			.or(just('{').to(Key::CurBrackL))
			.or(just('}').to(Key::CurBrackR))
			.or(just('[').to(Key::SquBrackL))
			.or(just(']').to(Key::SquBrackR))
			.or(just('<').to(Key::AngBrackL))
			.or(just('>').to(Key::AngBrackR))
			.or(just(':').to(Key::Colon))
			.or(just(';').to(Key::Semi))
			.or(just(',').to(Key::Comma))
			.or(just("...").to(Key::Unfold))
			.or(just('.').to(Key::Dot))
			.or(just("==").to(Key::Eq))
			.or(just("!=").to(Key::Neq))
			.or(just("=").to(Key::Assign))
	}

	pub fn tok(self) -> Tok {
		Tok::Key(self)
	}
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Tok {
	Id(String),
	Str(String),
	Num(String),
	Bool(bool),
	Key(Key),
	Let,
	Fn,
	Type,
	If,
	Else,
	Loop,
	While,
	For,
	In,
	Use,
	Pub,
	And,
	Or,
	Is,
}

impl Display for Tok {
	fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
		write!(f, "Token")
	}
}

impl Tok {
	pub fn parser() -> impl AnyParser<char, (Tok, Span)> {
		let key = Key::parser().map(Self::Key);

		let integer = int(10);
		let float = int(10)
			.or_not()
			.map(|op| op.unwrap_or("".into()))
			.then(just('.').then(int(10)))
			.map(|(int, (dot, rest))| format!("{int}{dot}{rest}"));
		let num = integer.or(float).map(Self::Num);

		let str = just('"')
			.not()
			.repeated()
			.delimited_by(just('"'), just('"'))
			.map(|e| e.iter().collect())
			.map(Self::Str);

		let id = ident().map(|id: String| match id.as_str() {
			"true" => Self::Bool(true),
			"false" => Self::Bool(false),
			"if" => Self::If,
			"else" => Self::Else,
			"let" => Self::Let,
			"fn" => Self::Fn,
			"type" => Self::Type,
			"loop" => Self::Loop,
			"while" => Self::While,
			"for" => Self::For,
			"in" => Self::In,
			"use" => Self::Use,
			"pub" => Self::Pub,
			"and" => Self::And,
			"or" => Self::Or,
			"is" => Self::Is,
			_ => Self::Id(id),
		});

		key.or(str)
			.or(num)
			.or(id)
			.then_ignore(whitespace())
			.map_with_span(|tok, span| (tok, span))
	}

	pub fn parse_all() -> impl AnyParser<char, Vec<(Tok, Span)>> {
		let toks = Self::parser().then_ignore(whitespace()).repeated();
		whitespace().ignore_then(toks).then_ignore(end()).collect()
	}
}
